# -*- coding: utf-8 -*-
"""
    Module: E5A camac data analysis
    File:   e5_DB_fio.py
    Descr.: E5A解析: Run DB file I/O 
    Update: 210909 build start Ay
    Last:   221123 v.2.1 refactored
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import os
import glob
import numpy    as np
import pandas   as pd

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylibSbs         as ay

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#from    e5ana   import() 

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Run Log FIO
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

def RunLog_read(fn:str, *,
                fn_excel:str=None) -> object:
    """  Read CAMAC Run-Log file into pd.DataFrame
    fn:         str     file name path (*.log)
    <kwd args>
    fn_excel:   str     if not None, save pd.DataFrame as Excel file
    <Ret.>
    df:         pd.DataFrame
    <cf.J>
        incotblm:/usr/babirl/devtool/prinfo2.c は、TAB 区切り text で出力している。
Run Log File format:
fname = '_' + ExpID + '.log' as defa.
['fnH']     ['fnR'] ['TmSta']  ['TmEnd']    ['Date']   ['cmtS']               ['cmtE'] ''

plhv01-	    0001	09:03:27	09:03:59	29-Oct-21	plhv01-01 700V 500um, 25um	30sec	
plhv01-	    0002	09:04:49	09:05:21	29-Oct-21	plhv01-02 800V     30sec
  ...	
pre-	    0001	20:59:09	21:55:00	28-Oct-21	pre-01 SSD pulser C->Ar Gain ni shita	end sB2 hikui ??	
scnAtt01-	0001	10:32:36	10:33:06	29-Oct-21	scanAtt001 Mul 2000 	scnAtt Run001 end	
  ...	
scnEDssd01-	0001	05:29:29	05:33:25	30-Oct-21	scanED001    0.0 um 	ssdRun001 end	
  ...	
  ['fn']    = ['fnH'] + ['fnR']
  ['runNo'] = ['fnR']
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(os.path.exists(fn) != True):
        msg = '#@{:s}: canceled as fn[{:s}] is not exist'.format(
            __name__,fn)
        raise FileNotFoundError(msg)
    df0 = pd.read_csv(fn,header=None,sep='\t',engine='python',dtype=object)
    #df0 = df0.drop(df0.index[-1])           # 最終行  を削除
    df0.columns = ['fnH','fnR','TmSta','TmEnd',
                   'Date','cmtS','cmtE','']

    """  fnH = NaN の行を削除 """
    df0.dropna(subset=['fnH'],inplace=True)

    df0.insert(0,'fnm',0)
    df0.insert(1,'runNo',0)
    df0['fnm'] = df0['fnH']+df0['fnR']
    df0['runNo'] = df0['fnR'].astype(int)
    #df0['runNo'].astype(int)
    df1 = df0.set_index('fnm',drop=False)
    if(fn_excel is not None):
        df1.to_excel(fn_excel,sheet_name='_runlog') 
    return( df1 )

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        super class RunLog
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class RunLog( object ):
    """ E5_Rawdata RunLog
Params. for constructor:
    fpath_log:  str     path+filename of run-log
    kfnH:       str     key-fnH eg) 'scnEDssd01-'
Attributes:
    see ._myclear()
HowTo: eg.)
    rLg = RunLog(fpath)             # read file into rLg.df
    rLg.set_key_fnH('scnEDssd')     # selection by key, result = rLg.dfk
    print( rLg.getSelected_all() )  # show selected results
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _myclear(self):
        self.df     = None      # pd.DF runlog(org) as pd.DataFrame
        self.dfk    = None      # pd.DF runlog(kfnH selected)
        self.kfnH   = ''        # str   current fnH key
        
    def __init__(self, fpath_log:str, kfnH:str=None):
        """  Constructor
        fpath_log:  str     path+filename of run-log
        kfnH:       str     find key-fnH        eg) 'scnEDssd01'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._myclear()
        df0 = RunLog_read(fpath_log)
        self.df = df0
        if(kfnH is not None):
            self._set_key(df0,kfnH)
        
    def set_key_fnH(self, kfnH:str):
        """  Set key to find files
        kfnH:       str     find key-fnH        eg) 'scnEDssd01'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        df0 = self.df
        self._set_key(df0,kfnH)

    def _set_key(self, df0, kfnH):
        self.kfnH = kfnH
        df1 = df0.set_index('fnH',drop=False)
        key = kfnH + '-'
        df2 = df1.loc[key]      # df2 is DataFrame
        #print('### df2=',type(df2),df2)
        self.dfk = df2.set_index('runNo',drop=False)        
                
    def _run_info_txt(self, dfloc, mdFmt:int=0):
        """  run info to text
        dfloc:  DataFrame   df.loc() selected 
        mdFmt:  int         Ret. txt format
            = 0 :   ['fnm'] : ['cmtS']
            = 1 :   not yet implemented
        <Ret.> str
        eg) 'scnEDssd01-0001: 0.0 um'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(mdFmt == 0):
            tx1 = dfloc['cmtS']
            tx2 = dfloc['fnm']+': '+' '.join(tx1.split()[1:])
        return(tx2)        

    def get_by_fnm(self, fnm:str , *,
                   mdFmt:int=0):
        """  Get run info by file name
        fnm:    str     run file name   eg) 'scnEDssd01-0001'
        mdFmt:  int     Ret. txt format ->see) ._run_info_txt()
        <Ret.> str
        eg) 'scnEDssd01-0001: 0.0 um'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        df0 = self.df
        df1 = df0.set_index('fnm',drop=False)
        df2 = df1.loc[fnm]      # df2 is Series
        # print('### df2=',type(df2),df2)
        return( self._run_info_txt(df2,mdFmt=mdFmt) )        
        
    def getSelected_by_RunNo(self, runNo:int, *,
                    mdFmt:int=0):
        """  Get run info by runNo from selected runs
        runno:  int     runNo to find
        mdFmt:  int     Ret. txt format ->see) ._run_info_txt()
        <Ret.> str
        eg) 'scnEDssd01-0001: 0.0 um'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        loc = self.dfk.loc[runNo]
        return( self._run_info_txt(loc,mdFmt=mdFmt) )        

    def getSelected_all(self, *,
                    mdFmt:int=0):
        """  Get info of all-selected runs
        mdFmt:  int     Ret. txt format ->see) ._run_info_txt()
        <Ret.> list(str)
        eg) ['scnEDssd01-0001: 0.0 um',.. ]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        rnoL = np.array(self.dfk['runNo'])
        txL =[]
        for rno in rnoL:
            txL.append(self.getSelected_by_RunNo(rno,mdFmt=mdFmt))
        return( txL )

    def getSelected_RunNo(self):
        """  Get RunNo of selected runs
        <Ret.> list(int)
        eg) [1,2,.. ]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( list(self.dfk['runNo']) )

    def getSelected_fnm(self):
        """  Get file name of selected runs
        <Ret.> list(str)
        eg) ['scnEDssd01-0001',.. ]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( list(self.dfk['fnm']) )


""".----.---.---.---.---.---.---.---.---.---.---.---.---
        helper func. for RunLog()
---.----.---.---.---.---.---.---.---.---.---.---.---.---"""

def get_RunLog(ExpID:str, dir1:str, dir2:str='_RawDmp',
               fn:str=None, kfnH:str=None, swfn1:bool=False ) -> object:
    """  get Run Log                    eg)
    ExpID:      str     ExpID           '2201Kr'
    dir1:       str     DB dir          'Q:' or '../'
    dir2:       str     RawDmp dir      defa) '_RawDmp'
    fn:         str     RunLog fname    defa) '_'+ExpID+'.log'
        ie) path = dir1 + ExpID + dir2 + fn
                 = Q:2201Kr/_RawDmp/_2201Kr.log
    kfnH:       str     find key-fnH    'scnEDssd01'
        ie) find files = kfnH + '_' + '*'
    <Ret.>
    rLg:        runLog()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    fn   = '_'+ExpID+'.log' if(fn is None) else fn
    pth1 = os.path.join(dir1,ExpID)
    pth2 = os.path.join(pth1,dir2)
    fn   = os.path.join(pth2,fn)
    #print('### RunLog fn =', fn)
    rLg  = RunLog(fn,kfnH)
    return( rLg )


""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Run DB FIO
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

def get_Data_fnL(ExpID:str, dir1:str, dir2:str='_RawDmp', 
                 fnH:str=None) -> list:
    """  get Data fname list                eg)
    ExpID:      str     ExpID               '2201Kr'
    dir1:       str     DB dir              'Q:' or '../'
    dir2:       str     RawDmp dir          defa) '_RawDmp'
    fnH:        str     Data fn Wild Card   'scnEDssd01' 
        ie) path = dir1 + ExpID + dir2 + fnH
                 = Q:2201Kr/_RawDmp/scnEDssd01-*.dmp
    <Ret.>
    fnL:    list(str)   file path list
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(fnH is None):
        msg = '#@{:s}:Err canceled as fnH[{:s}] is not specified'.format(
            __name__,fnH)
        raise FileNotFoundError(msg)
    pth1 = os.path.join(dir1,ExpID)
    pth2 = os.path.join(pth1,dir2)
    fnWC = os.path.join(pth2,fnH+'*.dmp')
    #print('### fnWC=',fnWC)
    fnL  = glob.glob(fnWC)
    if(len(fnL) < 1):
        msg = '#@{:s}:Err canceled as fnWC[{:s}] is not found'.format(
            __name__,fnWC)
        raise FileNotFoundError(msg)        
    return( fnL )

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        Run Data DB helper func
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
def get_RunDB_ix(rDB:list, ExpID:str) -> list:
    """  get Run_DB[]  indices
    rDB:    list    runDB params    def. in e5DB_*.py
    ExpID:  str     Experiment ID   eg) '2210Kr'
    <Ret.>
    [int]   ix of Run_DB[ix] where ExpID is matched
     ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    L=list()
    for i,prm in enumerate(rDB):
        expID,*_ = prm
        if(expID == ExpID):
            L.append(i)
    if(len(L) < 1):
        msg = '#@{:s}:Err canceled as ExpID[{:s}] is not found'.format(
            __name__,ExpID)
        raise FileNotFoundError(msg)        
    return( L )        

""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest01 = False    # RunLog_read(w/save_to_Excel), runLog()
    SwTest02 = True     # rLg= get_RunLog(), rLg.get_Selected*()
    SwTest03 = False    # rLg= rLg.get_by_fnm()

    SwTest11 = False    # get_Data_fnL()
    SwTest12 = False    # get_Data_fnL()
    """-.---.---.---.---.---.---.---.---.---.---.---.---"""

    """ Work param   ---.---.---.---.---.---.---.---.---"""
    #DB_dir1 = 'Q:'                  # mount ~/_pyE5A-ExpData/ as Q: drive
    DB_dir1 = '../'                 # test data DB dir
    ExpID = '0000Test'
    DB_dir2 = '_RawDmp'             # dump file dir
    DB_fnDat= 'scnEDssd01'          # CAMAC dump file Head name
    
    # CAMAC Data file name list : *.dmp 
    #dfnL = get_Data_fnL(DB_dir1, DB_dir2, DB_dir3, DB_fnDat)
    #print('### dfnL=',dfnL)

    """ test01 : RunLog_read(w/save_to_Excel), runLog()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest01):
        fn = '../0000Test/_RawDmp/_0000Test.log'
        df = RunLog_read(fn,fn_excel='_0000Test.xlsx')      
        # rLg = runLog(fn,'scnEDssd01-')    # read w/ kfnH

        rLg = RunLog(fn)                    # make instance first
        rLg.set_key_fnH(DB_fnDat)           # then, set kfnH
        print(rLg.getSelected_all())
        print('rLg.df : all data =')
        print(rLg.df)
        print('rLg.dfk : selected by key_fnH =')
        print(rLg.dfk)       

    """ test02 : rLg= get_RunLog(), rLg.get_Selected*()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest02):
        rLg = get_RunLog(ExpID, DB_dir1, DB_dir2 )
        rLg.set_key_fnH(DB_fnDat)           # set kfnH
        print('R#01 =',rLg.getSelected_by_RunNo(1))
        print('RunNo=',rLg.getSelected_RunNo())
        print('Fnm  =',rLg.getSelected_fnm())
        print(rLg.getSelected_all())
 
    """ test03 : rLg.get_by_fnm()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest03):
        rLg = get_RunLog(ExpID, DB_dir1, DB_dir2)
        DB_fn = 'pre-0001'
        tx = rLg.get_by_fnm(DB_fn)
        print(tx)

    """ test11 : get_Data_fnL()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest11):
        dfnL = get_Data_fnL(ExpID, DB_dir1, DB_dir2, DB_fnDat)
        print('dfnL=',dfnL)
        
    """ test12 : get_RunDB_ix()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest12):
        from e5ana  import(rDB_scnEDssd)
        
        ExpID = '2210Ar'
        ixL = get_RunDB_ix(rDB_scnEDssd,ExpID)
        print('ExpID= ',ExpID,' ixL=',ixL)
        for ix in ixL:
            print('[',ix,']=')
            print(rDB_scnEDssd[ix])
      
    print('@@main: -- Fin ---')     # 以上です。

""" ========================================================
        ref) Numbered ref
    ========================================================
r#01) nkmk pandasで任意の位置の値を取得・変更するat, iat, loc, iloc
    https://note.nkmk.me/python-pandas-at-iat-loc-iloc/

    ========================================================
        ref) Others
    ======================================================== 

------------------------------------------------------------------
"""
