# -*- coding: utf-8 -*-
"""
    Module: E5A data analysis
    File:   e5ana_Gaf.py
    Descr.: E5A解析: Gaf-film analysys
    Update: 210710 build start Ay
            210725 v1.0 
            210825 v1.1 
                add minor-tick grid, HowTo; 
                fix ban-wid plot, get_band(); 
                chg oR -> oG で解析
            210825 v1.1.1
                aylib.plxy() 不使用
    Last:   230223 

ref) 神原 Gaf報告_生物C,Ar,Fe_HDV2,EBT3_190417.pdf
      カタログ値  :  神原実測(C,Ar,Fe)
 HDV2:10-1000Gy: R:max 100Gy, ODovf 0.4 G:200, 0.5 B:50-300, 0.2
 EBT3: 0.1-10Gy: R:max   2Gy, ODovf 0.3 G: 10, 0.5 B:使えない

@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  numpy               as np
import  pandas              as pd
import  matplotlib.pyplot   as plt
plt.style.use('default')
#plt.style.use('seaborn')
#plt.style.use('seaborn-whitegrid')
#plt.style.use('bmh')
import  matplotlib.patches  as patches
from    matplotlib.colors   import  Normalize

#from    scipy               import  ndimage   # use .center_of_mass()
from    scipy.ndimage        import  center_of_mass
import  io  # use BytesIO()

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    aypylibSbs  import  (fitGaRoi)

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    gaf         import  (rdGafScan1,rdGafScan6,rdGafScan6_Kam)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class AnaOD
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

class anaOD( object ):
    """ Analize Optical Density(O.D.) : ndarray
Params. for constructor:
        odC:  ndarray           Optical density of one color band
        xyw:  tuple(int,int)    [mm] odC x,y-size in mm
Attributes:
        see ._myclear()
HowTo:
    scn = rdGafScan6('gaf6.bmp',dpi=75) # 6枚貼用紙読取り,解析
    # 注) crop画像がズレていると、以下の解析がうまくできません。
    gaf = anaOD(scn.odR[0])     # OD_Red[0] を defa値で解析
    if( Wobbler = ON 画像(ban-x,y 解析)の場合 ):
        gaf.plotXY()                    # 解析結果を XY plot
        gaf.show()                      # 主要数値の表示
        gaf.save_fig('a1.png')           # XY plot を .png で保存
        # 解析パラメータの調整
        gaf.set_ticker(5)               # XY目盛 5mm 間隔
        gaf.set_band('x',(-10,10,5,2))  # ban-x (start,end,step,band幅)[mm]
        gaf.set_odBG(((20,20),(25,25))) # BG lev 評価位置 (x1,y1)-(x2,y2)
        gaf.plotXY(odMin=0.1,odMax=0.5) # OD値範囲を指定して XY plot
    if( Wobbler = OFF 画像(ban-x,y 解析無し、Gaussian Fit)の場合 ):
        gaf.plotXY(plavr=False)         # ban-xy 表示無しで XY plot
        (xm,xsg,xa,xR2),(ym,ysg,ya,yR2) \
            = gaf.plotFitGa()           # OD分布の重心で X,Y方向に Gauss Fit
        gaf.save_fig('a2.png')          # XY plot を .png で保存 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _myclear(self):
        # optical density data array
        self.odC    = None            # ndarray O.D. array of one color
        self.nx,  self.ny = 0,0       # [pix] odC x,y-size  in [pixel]
        self.xw,  self.yw = 0,0       # [mm]  odC x,y-size  in [mm]
        self.xpix,self.ypix=None,None # [pix] odC x,y-pixel index
        self.xp2m,self.yp2m=None,None # [mm]  odC x,y-pixel index -> mm
        self.odBG   = 0.0             # odC[] BG  level
        self.odBGbox= None            # [pix] '' sampling box (x1,y1),(x2,y2))
        self.odMax  = 0.0             # odC[] Max value
        self.odCMp  = (0,0)           # odC[] Center-of-Mass in [pix]
        self.odCMm  = (0,0)           #      ''              in [mm]
        self.odCMban= (None,None)     # odC[] CM xy-band 1D-ndarray
        self.odCMavr= (0,0)           # odC[] CM xy-band average
        # for plot
        self.fig = None                       # figure
        self.wid, self.hgt = None,None        # axs wid/hight
        self.axs = None                       # axs[,]
        self.xtick,  self.ytick = None,None   # [mm] xy-plot ticker
        self.xtickMi,self.ytickMi = None,None #  minor ticker for grid
        self.xlbl,   self.ylbl  = None,None   #      xy-plot ticker label
        self.xbanprm,self.ybanprm = None,None # xy-band param (sta,end,stp,wid)
        self.xbanWm, self.ybanWm = 0,0        # [mm]  xy-band width in [mm]
        self.xbanWp, self.ybanWp = 0,0        # [pix]  ''           in [pix]
        self.xbanPm, self.ybanPm = None,None  # [mm]  xy-band center position
        self.xbanPp, self.ybanPp = None,None  # [pix]  ''
        self.xbanV,  self.ybanV  = None,None  # [od]  xy-band O.D. value
        self.xbanLc, self.ybanLc = None,None  # [color] '' line color 'CN' format
        self.xbanLw, self.ybanLw = None,None  # [pix]   '' line width      
        # Gauss fit result
        self.xfitV,  self.yfitV  = None,None  # [od]  xy-Gauss fit O.D. value
        self.xfitR,  self.yfitR  = None,None  # fit result (m,sg,a,R2)
         
    def __init__(self, odC:np.ndarray, xyw=(85,85)):
        """  Constructor
        odC:  ndarray           Optical density of one color band
        xyw:  tuple(int,int)    [mm] odC x,y-size in mm
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._myclear()
        self.odC = odC
        self.nx,self.ny = odC.shape
        self.xw,self.yw = xyw
        self.xpix = np.arange(0,self.nx)        # pixcel index list
        self.ypix = np.arange(0,self.ny)
        self.xp2m = self._pix2mm(self.xpix,'x') # pixcel index -> mm 
        self.yp2m = self._pix2mm(self.ypix,'y')
        # Wob=ON/OFF 共通の解析　までやっておく
        self.set_ticker()
        self.set_band('x')
        self.set_band('y')
        self.set_odBG()
        self.set_odMax()
        self.get_odCM()
        self._odCM_banxy()
        self._set_fig()

    def _wn(self,ax):
        """  get wid[mm] & n[pix] of o.d.image
        ax: char            'x' or 'y' axis name
        (retern)    (wid,n)
        wid:    float   mm_width     of odC[] == self.xw|yw
        n:      int     pixel_width of odC[] == self.nx|ny 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        w = self.xw if ax == 'x' else self.yw 
        n = self.nx if ax == 'x' else self.ny
        return(w,n)

    def _pix2mm(self,pix,ax):
        """  pixel --> mm 
        pix: int or [int]       [pix] pos of odC[]
        ax:  char               'x' or 'y' axis name
        (retern)    mm
        mm:  float or [float]   [mm]  pos of odC[]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        w,n = self._wn(ax)
        mm = (w / n)*(pix - n/2) if ax == 'x' else \
             (w / n)*(n/2 - pix)        # imshow() の Y軸は逆なので
        return( mm )

    def _pix2mmXY(self,xp,yp):
        """  pixel (xp,yp) --> mm (xmm,ymm) 
        <Ret.>    (xmm,ymm)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        xm,ym = self._pix2mm(xp,'x'),self._pix2mm(yp,'y')
        if((xm < -self.xw/2) or (xm > self.xw/2) or \
           (ym < -self.yw/2) or (ym > self.yw/2)):
            raise ValueError('xp or yp out of range')
        return(xm,ym)

    def _mm2pix(self,mm,ax):
        """  mm -> pixel 
        mm:  float or [float]   [mm] pos of odC[]
        ax:  char               'x' or 'y' axis name
        (retern)    pix
        pix: int or [int]       [mm]  pos of odC[]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        w,n = self._wn(ax)
        px = (n / w)*mm + n/2 if ax == 'x' else \
             n/2 - (n / w)*mm           # imshow() の Y軸は逆なので
        return( px.astype('int') if isinstance(mm,np.ndarray) \
                else int(px))

    def _mm2pixXY(self,xm,ym):
        """  mm (xmm,ymm) --> pixel (xp,yp)
        <Ret.>    (xp,yp)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        xp,yp = self._mm2pix(xm,'x'),self._mm2pix(ym,'y')
        if((xp < 0) or (xm > self.nx) or \
           (yp < 0) or (ym > self.ny)):
            raise ValueError('xm or ym out of range')
        return(xp,yp)

    def _pix2mmMag(self,ax):
        """  Magnification factor of pixel -> mm  
        ax:  char       'x' or 'y' axis name
        (retern)    mag
        mag: float      mag.factor == wid[mm]/n[pix]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        w,n = self._wn(ax)
        return( w/n )

    def _mm2pixMag(self,ax):
        """  Magnification factor of mm -> pixel  
        ax:  char       'x' or 'y' axis name
        (retern)    mag
        mag: float      mag.factor == n[pix]/wid[mm]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        w,n = self._wn(ax)
        return( n/w )

    def get_box(self, boxmm) -> np.ndarray:
        """  get box of odC[]
        boxmm:  (xm1,ym1),(xm2,ym2) [mm] box area in mm
        (retern)    (od_box,boxpix)
        od_box: 2D-ndarray          box of odC[]
        boxpix: (xp1,yp1),(xp2,yp2) [pix] box area in pixcel
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        (xm1,ym1),(xm2,ym2) = boxmm
        xp1,yp1 = self._mm2pixXY(xm1,ym1)
        xp2,yp2 = self._mm2pixXY(xm2,ym2)
        if(xp1 > xp2): xp1,xp2 = xp2,xp1       
        if(yp1 > yp2): yp1,yp2 = yp2,yp1       
        od1 = self.odC
        od2 = od1[xp1:xp2,yp1:yp2]
        return( od2, ((xp1,yp1),(xp2,yp2)) )

    def get_band(self, ax, banmm):
        """  get band of odC[]
        ax:     char        'x' or 'y' axis name
        banmm:  (mm1,mm2)   [mm]  (band_start,_end) in mm
        (retern)    (od_ban,boxpix)
        od_ban: 2D-ndarray  band of odC[]
        banpix: (px1,px2)   [pix] (band_start,_end) in pixcel
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        mm1,mm2 = banmm
        ax2 = 'x' if ax=='y' else 'y'   # 注) mm2pix変換の ax2 は x,y逆
        px1 = self._mm2pix(mm1,ax2)
        px2 = self._mm2pix(mm2,ax2)
        if(px1 > px2): px1,px2 = px2,px1       
        od1 = self.odC
        od2 = od1[px1:px2,:] if ax=='x' else od1[:,px1:px2].T
        #print('get_band',ax,banmm,px1,px2)
        return( od2, (px1,px2) )

    def get_bandxy(self, banpos=(0.,0.), banw=2.):
        """  make xy-band of odC[]
        banpos:     (xm,ym)     [mm] xy-band center pos.
        banw:                   [mm] band width
        <Ret.>    (banx,bany)
        banx,bany:  1D-ndarray  ban{x|y}.mean(axis=0)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        xcm,ycm = banpos
        hbw = banw/2
        x1m,x2m = xcm-hbw, xcm+hbw
        y1m,y2m = ycm-hbw, ycm+hbw
        bnx1,_ = self.get_band('x',(y1m,y2m))
        bny1,_ = self.get_band('y',(x1m,x2m))
        bnx2,bny2 = bnx1.mean(axis=0),bny1.mean(axis=0)
        return( bnx2,bny2 )

    def _band_avr(self, ban: np.ndarray, 
                  vmin:float = None, vmax:float = None, avrbias=0.2 ):
        """  calc  average( xy-band )
        ban:        1D-ndarray
        vmin,vmax:  float   if None, auto calc as ban{.min()|.max()}
        avrbias:    float   threshold = (max - min)*bias + min
        <Ret.>    avr
        avr:        float   average == ban[ban>thr].mean()
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(vmin is None): vmin=ban.min()
        if(vmax is None): vmax=ban.max()
        thr =(vmax - vmin)*avrbias+vmin
        ban1 = ban[ban > thr]
        avr  = ban1.mean()
        return( avr )

    def set_band(self, ax, banprm=(-15,15,5,3) ):
        """  set band plot of odC[]
        ax:     char    'x' or 'y' axis name
        banprm: int*4   [mm] (ban_start,_end,_step,_wid)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        ban_sta,ban_end,ban_stp,ban_wid = banprm

        def _ban(ax):
            banp = np.arange(ban_sta,ban_end+1,ban_stp)
            banv, banc = [],[]
            for i,mm0 in enumerate(banp):
                mm1,mm2 = mm0-ban_wid/2,mm0+ban_wid/2
                banod,_ = self.get_band(ax,(mm1,mm2)) 
                avr = banod.mean(axis=0)        # average in ban_wid
                banv.append(avr)
                banc.append('C{}'.format(i))    # line color
            # print('_ban',ax,banp,banc)
            return(banp,banv,banc)
        
        #print('set_band',ax,banprm)
        if(ax == 'x'):
            self.xbanprm= banprm
            self.xbanWm = ban_wid
            self.xbanWp = self._mm2pixMag('y')*ban_wid
            self.xbanPm, self.xbanV, self.xbanLc= _ban('x')
            self.xbanPp = self._mm2pix(self.xbanPm,'y')
            self.xbanLw = np.full(len(self.xbanPm),self.xbanWp,dtype=int)
        if(ax == 'y'):
            self.ybanprm= banprm
            self.ybanWm = ban_wid
            self.ybanWp = self._mm2pixMag('x')*ban_wid
            self.ybanPm, self.ybanV, self.ybanLc= _ban('y')    
            self.ybanPp = self._mm2pix(self.ybanPm,'x')
            self.ybanLw = np.full(len(self.ybanPm),self.ybanWp,dtype=int)

    def set_odBG(self,boxmm=((-2.5,31),(2.5,34)) ) -> float:
        """  set odC[] BG level from specified area 
        boxmm   (xm1,ym1),(xm2,ym2) [mm] box area which gives BG lev.
        <Ret.>    odBG
        odBG:   float(0.~1.0)   obtained BG level
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        od1, self.odBGbox = self.get_box(boxmm)
        self.odBG = od1.mean()
        return(self.odBG)

    def set_odMax(self,odmax:float=None ) -> float :
        """  set o.d. max level 
        odmax   float(0.~1.0)   if None, auto get from odC image
        <Ret.>    odMax
        odMax:  float(0.~1.0)   obtained max(odC[])
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(odmax is not None):
            self.odMax = odmax
        else:
            self.odMax = self.odC.max()
        return(self.odMax)

    def get_odCM(self, odbias:float = 0.2) -> tuple :
        """  get Center_of_Mass of odC[]  -> r#05)
        odbias      float       threshold = (Max - BG)*bias + BG
        <Ret.>    (xmm,ymm) 
        (xmm,ymm)   [mm]        obtained CM position
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        od1 = self.odC
        thr =(self.odMax - self.odBG)*odbias+self.odBG
        od2 = np.where(od1 < thr,0,od1)
        #yp,xp = ndimage.measurements.center_of_mass(od2)
        yp,xp = center_of_mass(od2)
        self.odCMp = (xp,yp)
        self.odCMm = (self._pix2mmXY(xp,yp))
        return( self.odCMm )

    def _odCM_banxy(self, banw=2., avrbias=0.2):
        """  calc   self.odCMavr
        banw:               [mm] band width
        avrbias:    float   threshold = (max - min)*bias + min
       ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        bnx,bny = self.get_bandxy(banpos=self.odCMm,banw=banw)
        avx = self._band_avr(bnx,
                vmin=self.odBG,vmax=self.odMax,avrbias=avrbias)
        avy = self._band_avr(bnx,
                vmin=self.odBG,vmax=self.odMax,avrbias=avrbias)
        self.odCMban = (bnx,bny)
        self.odCMavr = (avx,avy)
        
    def set_ticker(self, step=10):
        """  set xy-ticker & labe               -> r#06)
        tick_step: int     [mm] xy-plot ticker step
       ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def lbl_tick(npx, wmm, tcs):
            hw  = int(wmm / 2)
            htc = int(hw/tcs)*tcs
            lbl = np.arange(-htc,htc+1,tcs)
            tck = lbl * npx/wmm + npx/2
            lbl2= np.arange(-htc,htc+1,tcs/2)
            tckMi = lbl2 * npx/wmm + npx/2
            return(lbl,tck,tckMi)
        
        self.xlbl, self.xtick, self.xtickMi = \
            lbl_tick(self.nx,self.xw,step)
        self.ylbl, self.ytick, self.ytickMi = \
            lbl_tick(self.ny,self.yw,step)
        self.ytick   = np.flipud(self.ytick).copy() # imshow() の Y軸は逆なので
        self.ytickMi = np.flipud(self.ytickMi).copy()

    def _set_fig(self):
        """  set plot fig, axes
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        fig = plt.figure(figsize=(7,7))     # defa size=(6.4,4.8)
        hN,vN   = 2,2           # num.of plots 横_column:hN,縦_row :vN    
        wid,hgt = [1,2],[2,1]   # plot size ratio 横_wid, 縦_hgt
        gs  = fig.add_gridspec(ncols=hN, nrows=vN,
                    width_ratios=wid,height_ratios=hgt)
        axs = np.full((vN,hN),None)
        # sharex,y は　後からでは指定できない
        axs[0,0] = fig.add_subplot(gs[0,0],sharey=axs[0,1])                 
        axs[0,1] = fig.add_subplot(gs[0,1])
        axs[1,0] = fig.add_subplot(gs[1,0],sharex=axs[0,1]) 
        axs[1,1] = fig.add_subplot(gs[1,1])
        self.fig, self.axs = fig, axs
        self.wid, self.hgt = wid, hgt
       
    def plotXY(self, odMin:float=None, odMax:float=None, \
               plavr:bool=True, plavrV:int=10):
        """  plot : xy-image & xy-ban
        odMin:  float(0.0 ~ 1.0) if None, use self.odBG
        odMax:  float(0.0 ~ 1.0) if None, use self.odMax
        plavr:  bool             SW for plot avr_lines
        plavrV: int in [%]       avr +/- plavrV[%] 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        fig = self.fig
        ax = self.axs[1,0]
        ax.axis('off')        
        """ figure info """
        x0,y0,dy = 0.05, self.hgt[-1]/sum(self.hgt), 0.03
        fig.text(x0,y0-dy,  'Opt.Dens.')
        fig.text(x0,y0-dy*2,' BG ={:.2f}'.format(self.odBG))
        fig.text(x0,y0-dy*3,' max={:.2f}'.format(self.odMax))
        fig.text(x0,y0-dy*4, \
            ' center(x,y)=({:+.1f},{:+.1f})mm'.format(*self.odCMm))
        fig.text(x0,y0-dy*5, \
            ' avr(x,y)=({:.2f},{:.2f})od'.format(*self.odCMavr))
        """ plot xy-image  """
        ax = self.axs[0,1]
        vmin = self.odBG  if(odMin is None) else odMin
        vmax = self.odMax if(odMin is None) else odMax
        ax.imshow(self.odC,cmap="jet", norm=Normalize(vmin,vmax))
        ax.grid(axis='x', c='w', lw=0.2,which='both')
        ax.grid(axis='y', c='w', lw=0.2,which='both')
        if(True):  # thcker cnvert to [mm] Flase: image座標確認時
            ax.set_xticks(self.xtick)
            ax.set_xticks(self.xtickMi,minor=True)
            ax.set_xticklabels(self.xlbl)
            ax.set_yticks(self.ytick)
            ax.set_yticks(self.ytickMi,minor=True)
            ax.set_yticklabels(self.ylbl)
        if(False):  # thcker = OFF
            ax.axes.tick_params(axis='x',labelbottom=False)
            ax.axes.tick_params(axis='y',labelleft=False)
        """ over plot BG sampling position """
        (x1,y1),(x2,y2) = self.odBGbox
        w,h = x2-x1,y2-y1
        rec = patches.Rectangle((x1,y1),w,h,ec='w', fill=False)
        ax.add_patch(rec)                   # -> r#04)
        ax.text(x2,y1,'BG',c='w')
        """ over plot center_of_mass position  """
        xp,yp = self.odCMp
        dp = self._mm2pixMag('x') * 2       # cross = 2mm wid
        ax.hlines(yp,xp-dp,xp+dp,'b',lw=1)
        ax.vlines(xp,yp-dp,yp+dp,'b',lw=1)
        """ over plot xy-band position """
        px = self.xbanPp                    # -> r#03) hlines()版
        if(len(px)>0):
            pmax,pmin = int(self.nx*0.05),0
            ax.hlines(px,pmin,pmax,self.xbanLc,lw=self.xbanLw)
            ax.text(pmax,px[0],'ban-x',c='w',rotation='vertical')
        px = self.ybanPp
        if(len(px)>0):
            pmax,pmin = self.ny-1, int(self.ny*0.95)
            ax.vlines(px,pmin,pmax,self.ybanLc,lw=self.ybanLw)
            ax.text(px[-1],pmin,'ban-y',c='w')

        """ plot ban-x """
        # cmap = plt.get_cmap("tab10")      # -> r#01)
        ax = self.axs[1,1]
        ax.set_xmargin(0)       # X軸余白無し設定
        for i,yv in enumerate(self.xbanV):
            ax.plot(self.xpix,yv,lw=0.5,color=self.xbanLc[i])   
        ax.grid(axis='x', c='c', lw=0.2)
        ax.set_ylim(vmin,vmax)
        ax.set_xticks(self.xtick)
        ax.set_xticklabels(self.xlbl)
        ax.set_xlabel('X[mm]')
        ax.text(0.02,0.90,'ban-x',transform=ax.transAxes)

        """ plot ban-y """
        ax = self.axs[0,0]
        ax.set_ymargin(0)
        for i,yv in enumerate(self.ybanV):
        #   plot() には orientation='horizontal'が無い
            ax.plot(yv,self.ypix,lw=0.5,color=self.ybanLc[i])
        ax.set_xlim(vmin,vmax)
        ax.invert_xaxis()       # X,Y は、invert しているので
        ax.invert_yaxis()       # 以降の ax.text() で注意！ 特にprojY
        ax.grid(axis='y', c='c', lw=0.2)
        ax.set_yticks(self.ytick)
        ax.set_yticklabels(self.ylbl)
        ax.set_ylabel('Y[mm]')
        ax.text(0.02,0.95,'ban-y',transform=ax.transAxes)

        """ plot xy-avr"""
        if(not plavr): return
        msg = 'avr+/-{}%'.format(plavrV)
        avx,avy = self.odCMavr
        ax = self.axs[1,1]
        vmin,vmax = self.xtick[0],self.xtick[-1]
        p5p = (avx-self.odBG)*(1 + plavrV/100) + self.odBG
        m5p = (avx-self.odBG)*(1 - plavrV/100) + self.odBG
        ax.hlines(avx,vmin,vmax,'g','solid', lw=0.5)
        ax.hlines(p5p,vmin,vmax,'g','dashed',lw=0.3)
        ax.hlines(m5p,vmin,vmax,'g','dashed',lw=0.3)
        ax.text(vmax*0.8,avx,msg,c='g')
        ax = self.axs[0,0]
        vmin,vmax = self.ytick[-1],self.ytick[0]
        p5p = (avy-self.odBG)*(1 + plavrV/100) + self.odBG
        m5p = (avy-self.odBG)*(1 - plavrV/100) + self.odBG
        ax.vlines(avy,vmin,vmax,'g','solid', lw=0.5)
        ax.vlines(p5p,vmin,vmax,'g','dashed',lw=0.3)
        ax.vlines(m5p,vmin,vmax,'g','dashed',lw=0.3)
        ax.text(p5p,vmax*0.2,msg,c='g',rotation='vertical')
        fig.tight_layout()

    def plotFitGa(self, fnFit:callable=None):
        """  plot : Gaussian fit for odCMban
        fnFit:  callable    fit model func fnFit(x,xm,sg,a)
        <Ret.>
        fit_banx, fit_bany: Gauss fit result for odCMban-x,y
          where fit_ban = [m,sg,a,R2)]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def get_roi(x,y):
            """ get roi range for fitting """
            ymax= max(y) 
            thr = (ymax - self.odBG)* 0.5 + self.odBG  # < half.max
            wh = np.where(y >= thr)
            x1 = x[wh]
            roi1,roi2 = x1[0],x1[-1]
            #print('roi1,2=',roi1,roi2)
            return(roi1,roi2)

        def fn_Gaus_BG(x:list, xm:float, sgm:float, a:float):
            """ fitting func: Gaussian + const.BG= odBG """
            #print(BG)
            return( a * np.exp( - (x - xm)**2 / (2 * sgm**2))
                   + self.odBG)

        def fit_plot_ban(axes,axID,axV,binD,datD):
            """ fit & plot ban-x,y """
            roi1,roi2 = get_roi(binD,datD)
            if(axID == 'y'):
                roi1,roi2 = roi2,roi1
            [m,sg,a,R2],binF,fitF = fitGaRoi(
                    binD,datD, xroi1=roi1,xroi2=roi2,fnFit=_fnFit)
            #print('x:m,sg,a,R2=',xm,xsg,xa,xR2)
            datD2 = _fnFit(binD,m,sg,a)
            _plCol, _plLw = 'r', 1.     # fit plot color, Line wid
            tx1 = '(m,sg,a,R2)='
            tx2 = '({:+.2f},{:.2f},{:.2f},{:.3f})'.format(m,sg,a,R2)
            #px = - self.xw/3 if(axID == 'x') else  self.yw/3
            px = 0 if(axID == 'x') else self.ny
            dy = (self.odMax - self.odBG)
            py1,py2 = dy*0.8+self.odBG, dy*0.7+self.odBG
            if(axID == 'x'):
                axes.plot(axV,datD2,lw=_plLw,color=_plCol)
                axes.text(px,py1,tx1,c=_plCol)
                axes.text(px,py2,tx2,c=_plCol)
                self.xfitV = datD2
                self.xfitR = [m,sg,a,R2]
            else:
                axes.plot(datD2,axV,lw=_plLw,color=_plCol)
                axes.text(py1,px,tx1,c=_plCol,rotation='vertical')
                axes.text(py2,px,tx2,c=_plCol,rotation='vertical')
                self.yfitV = datD2
                self.yfitR = [m,sg,a,R2]
            return(m,sg,a,R2)
        
        _fnFit = fn_Gaus_BG if(fnFit is None) else fnFit
        x_dat,y_dat = self.odCMban
        """ plot ban-x,y """
        xm,xsg,xa,xR2 = fit_plot_ban(self.axs[1,1],'x',
                                     self.xpix,self.xp2m,x_dat)
        ym,ysg,ya,yR2 = fit_plot_ban(self.axs[0,0],'y',
                                     self.ypix,self.yp2m,y_dat)
        self.fig.tight_layout()
        return((xm,xsg,xa,xR2),(ym,ysg,ya,yR2))

    def save_fig(self, fn:str):
        """  save figure to file
        fn:     str     out file name eg) 'a.png'
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.fig.savefig(fn)

    def save_fig_IoByte(self):
        """  save figure to BytesIO object
        <Ret.>
        io.IoBytesIO    Byte stream
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        b = io.BytesIO()
        self.fig.savefig(b,format='PNG')
        b.seek(0)
        return( b )
        
    def show(self):
        """  show class attributes
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        print('odAna')
        print('.nx,ny  [pix] = {},{}'.format(self.nx,self.ny))
        print('.xw,yw  [mm]  = {},{}'.format(self.xw,self.yw))
        print('.odBG   [od]  = {:.2f}'.format(self.odBG),end='')
        (x1,y1),(x2,y2)=self.odBGbox
        print(' @ ({:+.1f},{:+.1f})-({:+.1f},{:+.1f})[mm]'.format(
            x1,y1,x2,y2))
        print('.odMax  [od]  = {:.2f}'.format(self.odMax))
        x1,y1=self.odCMm
        print('.odCMm  [mm]  =({:+.1f},{:+.1f})'.format(x1,y1))
        avx,avy=self.odCMavr
        print('.odCMavr[od]  =({:.2f},{:.2f})'.format(avx,avy))
        sta,end,stp,wid=self.xbanprm
        print('.xban(sta,end,stp,wid)=({:+.2f},{:+.2f},{:.2f},{:.2f})[mm]'.format(
            sta,end,stp,wid))
        sta,end,stp,wid=self.ybanprm
        print('.yban(sta,end,stp,wid)=({:+.2f},{:+.2f},{:.2f},{:.2f})[mm]'.format(
            sta,end,stp,wid))

    def mkdf_odC(self) -> object:
        """  make pd.DataFrame of .odC
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        df = pd.DataFrame(data=self.odC,
                index=self.yp2m, columns=self.xp2m)
        return(df)

    def mkdf_Atrb(self) -> object:
        """  make pd.DataFrame of anaOD attrubutions
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        anm =[  'nx,ny  [pix]',
                'xw,yw  [mm]',
                'od BG  [od]',
                'od Max [od]',
                'od center(x,y) [pix]',
                'od center(x,y) [mm]',
                'od Avr(x,y)    [od]',
                'G.fit  m(x,y)  [mm]',
                'G.fit sg(x,y)  [mm]',
                'G.fit  a(x,y)  [od]',
                'G.fit R2(x,y)' ]
        valx =[ self.nx,
                self.xw,
                round(self.odBG,2),
                round(self.odMax,2),
                int(self.odCMp[0]),
                round(self.odCMm[0],2),
                round(self.odCMavr[0],2) ]
        valy =[ self.ny,
                self.yw,
                '',
                '',
                int(self.odCMp[1]),
                round(self.odCMm[1],2),
                round(self.odCMavr[1],2) ]
        if(self.xfitR is not None):
            valx += self.xfitR
            valy += self.yfitR
        else:
            valx += ['','','','']
            valy += ['','','','']
        atrb = list(zip(anm,valx,valy))
        df1 = pd.DataFrame(atrb,columns=['Atr.nm','Atr.x','Atr.y'])
        df2 = self.mkdf_banxy('x')
        df3 = self.mkdf_banxy('y')
        df  = pd.concat([df1,df2,df3],axis=1)
        return(df)

    def mkdf_banxy(self,ax:str) -> object:
        """  make pd.DataFrame of .xbanV, .ybanV
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def _mkdf(p2m,banPm,banV,nm1,nm2,CMban,fitV,nm3,CMm):
            dfs = pd.DataFrame()
            sr = pd.Series(p2m,name=nm1)
            dfs = pd.concat([dfs,sr],axis=1)
            for i in range(len(banPm)):
                mm = banPm[i]
                # print('_mkdf',i,mm)
                sr1 = pd.Series(banV[i],name=nm2+'{:+.1f}'.format(mm))
                dfs = pd.concat([dfs,sr1],axis=1)
            sr1 = pd.Series(CMban,name=nm3+'{:+.1f}'.format(CMm))
            sr2 = pd.Series(CMban,name=nm3+'fit')
            dfs = pd.concat([dfs,sr1,sr2],axis=1)
            return( dfs )

        if(ax == 'x'): 
            df = _mkdf(self.xp2m,self.xbanPm,self.xbanV,'Xmm','Ymm=',
                       self.odCMban[0],self.xfitV,'Ycm=',self.odCMm[1])
        if(ax == 'y'): 
            df = _mkdf(self.yp2m,self.ybanPm,self.ybanV,'Ymm','Xmm=',
                       self.odCMban[1],self.yfitV,'Xcm=',self.odCMm[0])
        return(df)


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest01 =  True    # test Wob=ON  image
    SwTest02 =  False   # test Wob=OFF image    
    SwTest03 =  True    # test pd.DF save   
    SwTest11 =  False   # test as rdGafScn6_Kam()
    SwTest12 =  False   # test as rdGafScn1()
    """-.---.---.---.---.---.---.---.---.---.---.---.---"""
    fndir = '../0000Test/_GafData/'   # ExpData folder path

    """ Test file : 上記 SwTest に合わせて設定要  """
    fn_scn,dpi = fndir+'testGaf6ay_75dpi.bmp', 75
    #fn_scn,dpi = fndir+'testGaf6ay_300dpi.bmp', 300
    #fn_scn,dpi = fndir+'testGaf6km_75dpi.bmp', 75
    #fn_scn,dpi = fndir+'testGaf1_75dpi.bmp', 75

    if(SwTest01 or SwTest02):
        scn = rdGafScan6(fn_scn,dpi=dpi)
    if(SwTest11):
        scn = rdGafScan6_Kam(fn_scn,dpi=dpi)
    if(SwTest12):
        scn = rdGafScan1(fn_scn,dpi=dpi)

    """ test01,test11,test12  : test  Wob=ON  image
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest01 or SwTest11 or SwTest12):
        aOD = anaOD(scn.odG[5])         # use Wob=ON image  
        aOD.show()
        aOD.plotXY(odMin=0.1)         # check odBG pos.
        aOD.plotXY()
        """ # 以下 ban-xy をズラして確認したい場合
        aOD.set_ticker(10)              # set xy-thcker in mm
        xcm,ycm = aOD.odCMm             # Center_of_Mass pos
        aOD.set_band('x',(ycm-20,ycm+5,5,3)) # set xy-band_start,_end,_step,_wid in mm 
        aOD.set_band('y',(xcm,xcm+20,10,3))
        aOD.set_odBG(((20,20),(25,25))) # BG lev sampling box
        aOD.show()
        aOD.plotXY()
        """
        #aOD.save_fig('_anaOD1.png')
        #aOD.fig.savefig('_anaOD1.png')

    """ test02 : test  Wob=OFF image; Gauss-fit
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest02):
        aOD = anaOD(scn.odG[0])         # use Wob=OFF image  
        aOD.show()
        # aOD.plotXY(odMin=aOD.odBG-0.02)  
        # FIXME plotXY() ２回やると text位置が狂う！　ax.invert の影響
        # aOD.fig.savefig('_anaOD2a.png')
        
        aOD.set_ticker(10)              # set xy-thcker in mm
        xcm,ycm = aOD.odCMm             # Center_of_Mass pos
        aOD.set_band('x',(ycm,ycm,1,3)) # set xy-band_start,_end,_step,_wid in mm 
        aOD.set_band('y',(xcm,xcm,1,3))
        aOD.set_odBG(((20,20),(25,25))) # BG lev sampling box
        aOD.plotXY(odMin=0.15,odMax=0.8,plavr=False)
        BG = aOD.odBG
        #print('BG=',BG)

        def fn_gaus(x:list, xm:float, sgm:float, a:float):
            """ simple GaussFit w/o const BG """
            return( a * np.exp( - (x - xm)**2 / (2 * sgm**2)))

        def fn_gausBG(x:list, xm:float, sgm:float, a:float):
            """ This func is used in plotFitGa() """
            #print(BG)
            return( a * np.exp( - (x - xm)**2 / (2 * sgm**2)) + BG)

        if(False):
            pkfX,pkfY = aOD.plotFitGa(fnFit=fn_gaus)
        else:
            pkfX,pkfY = aOD.plotFitGa()
        xm,xsg,xa,xR2 = pkfX
        ym,ysg,ya,yR2 = pkfY
        print('GausFit banx:m,sg,a,R2=',end='')
        print('({:+.2f},{:.2f},{:.2f},{:.3f})'.format(xm,xsg,xa,xR2))
        print('        bany:m,sg,a,R2=',end='')
        print('({:+.2f},{:.2f},{:.2f},{:.3f})'.format(ym,ysg,ya,yR2))
        aOD.show()
        #aOD.fig.savefig('_anaOD2b.png')

    """ test03 : test  pd.DF save
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest03):
        df1 = aOD.mkdf_Atrb()
        df2 = aOD.mkdf_odC()
        with pd.ExcelWriter('anaOD.xlsx') as wr:
            df1.to_excel(wr,sheet_name='Atrb',float_format="%.3f")
            df2.to_excel(wr,sheet_name='odC', float_format="%.3f")

""" ========================================================
        ref) Numbered ref
    ========================================================
r#01) Qiita matplotlibでcolor cycleのN番目の色を指定するいくつかの方法
    https://qiita.com/skotaro/items/5c9893d186ccd31f459d

r#02) Qiita matplotlibのめっちゃまとめ 4.3. 矢印
     https://qiita.com/nkay/items/d1eb91e33b9d6469ef51
     
r#03) pystyle  matplotlib – 水平、垂直の線や帯を描画する方法
     https://pystyle.info/matplotlib-plot-horizontal-and-vertical-lines/

r#04) nkmk Matplotlibで円や長方形などの図形を描画
     https://note.nkmk.me/python-matplotlib-patches-circle-rectangle/

r#05) KWD =　ndarray の重心を求める
    scipy.ndimage.center_of_mass
    https://www.fixes.pub/program/463408.html

r#06) KWD = matplotlib thicker
    https://sabopy.com/py/matplotlib-12/
    https://pystyle.info/matplotlib-ticks-and-grid/
    https://www.delftstack.com/ja/howto/matplotlib/set-matplotlib-grid-interval/
    
    ========================================================
        ref) Others
    ======================================================== 

------------------------------------------------------------------
"""