# -*- coding: utf-8 -*-
"""
    Module: E5A data analysis
    File:   gaf_scan.py
    Descr.: E5A解析: Gaf-film analysys 
    Update: 210710 build start Ay
            210725 v1.0
            210825 v1.1 
                mod Cursor明るく 
                add HowTo, save_all as CSV
    Last:   220103 

@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  numpy               as np
from    PIL                 import Image,ImageDraw

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 
#from    aypylib.plot        import  (MPlot, Hist1D)

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        super class rdGafScan
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class rdGafScan( object ):
    """ Read Gaf Scan Image : rdGafScan*() common methods
Params. for constructor:
    non
Attributes:
    non
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _clearSP(self):
        pass
                  
    def __init__(self):
        """  Constructor
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._clearSP()

    def imCrop(self, im:Image, xp, yp, xpw, ypw) -> Image:
        """  Image crop                         -> r#03)
        im:         Image       PIL.Image
        xp,yp:      int         [pix] crop start pos
        xpw,ypw:    int         [pix] 
        <Ret.> 
        im.crop( (xp,yp)-(xp+xpw,yp+ypw) )
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        xp1,yp1 = xp+xpw, yp+ypw
        im1 = im.crop((xp,yp,xp1,yp1))
        return( im1 )

    def imRGB(self, im:Image, mode:int=0, fill:int=0):
        """  make RGB-band separated images     -> r#04,05) 
        im:         Image       PIL.Image
        mode = 0 :  imR,G,B[].mode == "L" 8-bit B/W
             = 1 :    == "RGB" 3x8-bit true color (R,0,0) (0,G,0) (0,0,B)
        fill:       int         when mode=1, fill data for each band 
        <Ret.> 
        [imR,imG,imB]: list of Image    RGB-band separated images
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(mode == 0):
            RGB = im.split()  # split img into individual bands
            return( RGB )
        # numpy ndarray に変換する方法しか分からなかった
        nda = np.array( im ) 
        ndR = nda.copy()
        ndR[:,:,(1,2)]= fill
        imR = Image.fromarray(ndR)
        ndG = nda.copy()
        ndG[:,:,(0,2)]= fill
        imG = Image.fromarray(ndG)
        ndB = nda.copy()
        ndB[:,:,(0,1)]= fill
        imB = Image.fromarray(ndB)    
        return( [imR,imG,imB] )

    def imDraw_Ellipse(self, im:Image, mxw:int, myw:int, fill:bool):
        """  Overlay Ellipse ; to chek Mask pos
        im:         Image       sample PIL.Image
        mxw,myw:    int         [pix] Mask-Ellipse size
            cf.) Ellipse is centered at the im
        fill:       bool        SW for fill ellipse area
        <Ret.> 
        imD:        Image       Draw image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        mode,size = im.mode, im.size
        xw,yw = size
        lw = int((xw + yw)/2*0.02)
        if(mxw > xw): mxw = xw
        if(myw > yw): myw = yw
        mask = Image.new(mode,size,color=0)
        draw = ImageDraw.Draw(mask)
        x1,y1 = int((xw-mxw)/2), int((yw-myw)/2)
        x2,y2 = x1+mxw, y1+myw
        if(fill):   draw.ellipse((x1,y1,x2,y2),fill=255)
        else:       draw.ellipse((x1,y1,x2,y2),width=lw)
        return( mask )

    def imDraw_Cross(self, im:Image, cxw:int, cyw:int, \
                         cxs:int=10, cys:int=10):
        """  Overlay Ellipse ; to chek Mask pos
        im:         Image       sample PIL.Image
        cxw,cyw:    int         [pix] Cross cursor size
        cxs,cys:    int         [pix] Cross cursor step
            cf.) Cross is centered at the im
        <Ret.> 
        imD:        Image       Draw image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        mode,size = im.mode, im.size
        xw,yw = size
        col,lwx,lwy = (255,255,255),int(xw*0.01),int(yw*0.01)
        dx,dy = int(xw*0.02), int(yw*0.02)
        if(cxw > xw): cxw = xw
        if(cyw > yw): cyw = yw
        mask = Image.new(mode,size,color=0)
        draw = ImageDraw.Draw(mask)
        """ darw x,y axis """
        x0,y0 = int(xw/2), int(yw/2)    # center pos
        x1,y1 = int(x0-cxw/2),int(y0-cyw/2)
        x2,y2 = int(x0+cxw/2),int(y0+cyw/2)
        draw.line(((x1,y0),(x2,y0)),fill=col,width=lwx*2)
        draw.line(((x0,y1),(x0,y2)),fill=col,width=lwy*2)
        """ darw x,y tick """
        nx = int(cxw/2/cxs)
        for i in range(nx):
            _dy = 2*dy if((i % 2) == 0) else dy
            x1,x2,y1,y2 = x0+i*cxs,x0-i*cxs,y0-_dy,y0+_dy
            draw.line(((x1,y1),(x1,y2)),fill=col,width=lwy)        
            draw.line(((x2,y1),(x2,y2)),fill=col,width=lwy)        
        ny = int(cyw/2/cys)
        for i in range(ny):
            _dx = 2*dx if((i % 2) == 0) else dx
            x1,x2,y1,y2 = x0-_dx,x0+_dx,y0+i*cys,y0-i*cys
            draw.line(((x1,y1),(x2,y1)),fill=col,width=lwx)        
            draw.line(((x1,y2),(x2,y2)),fill=col,width=lwx)        

        return( mask )

    def imOvl_CursorEllipse(self, im:Image, mxw:int, myw:int):
        """  Overlay Ellipse Cursor ; to chek Mask pos
        im:         Image       PIL.Image
        mxw,myw:    int         [pix] Mask-Ellipse size
            cf.) Ellipse is centered at the im
        <Ret.> 
        imM:        Image       Overlayed image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        im1 = im.copy()
        im2 = self.imDraw_Ellipse(im1, mxw, myw, fill=False)
        imM = Image.blend(im1, im2, alpha=0.2)  # alpha enhanse= 0:im1 1:im2
        return( imM )
        
    def imOvl_CursorCross(self, im:Image, cxw:int, cyw:int, \
                          cxs:int=10, cys:int=10):
        """  Overlay Ellipse Cursor ; to chek Mask pos
        im:         Image       PIL.Image
        cxw,cyw:    int         [pix] Cross cursor size
        cxs,cys:    int         [pix] Cross cursor step
            cf.) Cross is centered at the im
        <Ret.> 
        imM:        Image       Overlayed image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        im1 = im.copy()
        im2 = self.imDraw_Cross(im1, cxw, cyw, cxs, cys)
        imM = Image.blend(im1, im2, alpha=0.2)  # alpha enhanse= 0:im1 1:im2
        return( imM )

    def imOvl_Cursor(self, im:Image, mxw:int, myw:int, \
                          cxs:int=10, cys:int=10):
        """  Overlay Ellipse & Cross Cursor
        im:         Image       PIL.Image
        mxw,myw:    int         [pix] Ellipse size
        cxs,cys:    int         [pix] Cross cursor step
            cf.) Ellipse & Cross are centered at center of the im
        <Ret.> 
        imM:        Image       Overlayed image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        #mode,size = im.mode, im.size
        #blnk = Image.new(mode,size,color=255)
        im1 = im.copy()
        im2 = self.imDraw_Ellipse(im, mxw, myw, fill=False)
        im3 = self.imDraw_Cross(  im, mxw, myw, cxs, cys)
        #im4 = Image.blend(im2, im3, alpha=0.5)  # 薄くなる
        np2,np3 = np.array(im2), np.array(im3)
        np4 = np2 + np3         # 単純加算にして  im4=im2+im3 は不可なので
        np4 = np4.clip(0,255)   # 255 で clip
        im4 =Image.fromarray(np4.astype(np.uint8))
        im5 = Image.blend(im1, im4, alpha=0.2)  # alpha enhanse= 0:im1 1:im2
        return( im5 )
        
    def imMask_Ellipse(self, im:Image, mxw:int, myw:int):
        """  Composite Ellipse-Mask                 ->r#02)
        im:         Image       PIL.Image
        mxw,myw:    int         [pix] Mask-Ellipse size
            cf.) Ellipse is centered at the im
        <Ret.> 
        imM:        Image       Masked image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        mode,size = im.mode, im.size
        blnk = Image.new(mode,size,color=255) # color>0 後で od計算するので
        mask = self.imDraw_Ellipse(im, mxw, myw, fill=True)
        imM = Image.composite(im, blnk, mask)
        return( imM )

    def imOD(self, im:Image):
        """  Convert to Optical Density
        im:     Image       PIL.Image
        <Ret.> 
        od:     2D-ndarray  Optical Density (O.D.) = log10(255/im)
        odim:   Image       Image format of od
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        nda1= np.array( im, dtype=np.float )
        nda2= 255./nda1
        od  = np.log10(nda2)
        od1 = od * 255.0        # norm. OD=1.0 as color = 255
        od2 = np.array(od1,dtype=np.uint8)
        im1 = Image.fromarray(od2)
        return(od,im1)
         

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class rdGafScan6
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class rdGafScan6( rdGafScan ):
    """ Read Gaf Scan Image (E5A用 Gaf□85x85 6枚貼付け用紙の読取り)
Params. for constructor:
    fnScn:  str         Gaf scan image file name
    dpi:    int         [dpi] defa=75 
    ana:    bool        SW for scnAna_all() 
Attributes:
    see ._myclear()
HowTo:
    scn = rdGafScan6('gaf6.bmp',dpi=75) # 6枚貼用紙読取り, defa値で解析
    scn.show_all(mode=2,plt=True)       # show cropped images
    if( 6分割画像の切取りがズレていたら ):    # crop位置の補正をする
        sft = [None,None,(-4,2),(3,2),(-4,5),(2,5)]   # (dx,dy) x 6画面分
        for i,_sft in enumerate(sft):
            if(_sft is None): continue
            dx,dy=_sft
            scn.scnShift(i,dx,dy)   # crop 位置補正
        scn.scnAna_all()            # 再解析
        scn.show_all(mode=2,plt=True)
    scn.save_all(modes=[1,2,11,21,31])   # 6画像(.png)の保存
    # 続いて OD image の解析 --> see) anaOD()
    # gaf = anaOD(scn.odR[0])
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    def _myclear(self):
        n = 6 
        # Scanner params.
        self.dpi    = 0                 # [dpi] scan resolution N/A now
        self.x0     = [0,0]             # [pix] upper left pos.X 2列
        self.y0     = [0,0,0]           #    ''            pos.Y 3段
        self.xw,self.yw   = 0,0         # [pix] X,Ywidth
        self.xmw,self.ymw = 0,0         # [pix] Mask X,Ywidth
        self.xycp   = [0,0]*n           # [pix] Crop pos (xcp,ycp)
        # separated images
        self.im0    = None              # org image (incl. 6 images)
        self.nim    = n
        self.im     = [None]*n          # Image.mode='RGB' separated images    
        self.imCur  = [None]*n          # 'RGB' Overlay Ellipse Cursor    
        self.imR,self.imG,self.imB \
          = [None]*n,[None]*n,[None]*n  # 'L' R,G,B bands
        self.imRm,self.imGm,self.imBm \
          = [None]*n,[None]*n,[None]*n  # 'L' Masked img
        # Optical Density of RGB
        self.odR,self.odG,self.odB \
          = [None]*n,[None]*n,[None]*n  # ndarray.size = xw,yw
        self.imRod,self.imGod,self.imBod \
          = [None]*n,[None]*n,[None]*n  # Image.mode = 'L' 
                  
    def __init__(self, fnScn:str, dpi=75, ana=True):
        """  Constructor
        fnScn:  str         Gaf scan image file name
        dpi:    int         [dpi] defa=75 
            cf.) now support dpi=75,300 only
        ana:    bool        SW for scnAna_all() 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._myclear()
        super().__init__()
        self.im0 = Image.open(fnScn)
        self.dpi = dpi
        """ w10:ペイントで表示してカーソル位置で確認する！ """
        if(dpi == 75):
            self.x0     = [40,332]
            self.y0     = [52,326,602]
            self.xw,self.yw =   250,250 
            self.xmw,self.ymw = 210,210
        if(dpi == 300):
            mag= 300./75.
            self.x0     = [150,1325]
            self.y0     = [200,1300,2400]
            self.xw,self.yw =   250*mag,250*mag
            self.xmw,self.ymw = 210*mag,210*mag
        self._imCropPrep()
        if(ana):
            self.scnAna_all()

    def scnAna_all(self):
        """  Analize scan image 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self.imSep_all()
        self.imOvl_all()
        self.imRGB_all(mode=0)
        self.imMask_all()
        self.imOD_all() 

    def _imCropPrep(self):
        """  im0 crop pos calc
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def crpos(ix, iy) -> Image:
            xp,yp = self.x0[ix], self.y0[iy]
            return( xp,yp )

        self.xycp[0] = crpos(0,0)
        self.xycp[1] = crpos(1,0)
        self.xycp[2] = crpos(0,1)
        self.xycp[3] = crpos(1,1)
        self.xycp[4] = crpos(0,2)
        self.xycp[5] = crpos(1,2)

    def scnShift(self,imNo:int, dxm:int=0, dym:int=0):
        """  Shift x0[],y0[] separate position
        imNo:   int           im[] index 
        dxm:    int     [mm] shift in x-direction
        dym:    int     [mm]       in y-direction   
            cf.) after this, need to call scnAna_all()
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if((imNo < 0) or (imNo >= self.nim)):
            raise ValueError('imNo={} out of range'.format(imNo))
        dxp,dyp = int(self.dpi/25.4*dxm),int(self.dpi/25.4*dym)
        xp,yp = self.xycp[imNo]
        self.xycp[imNo] = (xp+dxp,yp-dyp) # Yは符号逆転; Plot表示はY軸が逆なので

    def scnShift_all(self,dxm:int=0, dym:int=0):
        """  Shift x0[],y0[] separate position
        dxm:    int     [mm] shift in x-direction
        dym:    int     [mm]       in y-direction        
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        dxp,dyp = int(self.dpi/25.4*dxm),int(self.dpi/25.4*dym)
        self.x0[0] += dxp
        self.x0[1] += dxp
        self.y0[0] += dyp
        self.y0[1] += dyp
        self.y0[2] += dyp
        self._imCropPrep()
        self.scnAna_all()

    def imSep_all(self):
        """  Separate im0 into 6 sections
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        for i in range(self.nim):
            xp,yp = self.xycp[i]
            self.im[i] = self.imCrop(self.im0,xp,yp,self.xw,self.yw)

    def imOvl_all(self, tick:int=5):
        """  Overlay Ellipse Cursor ; to chek Mask pos
        tick:   int     [mm] xy-axis tick step
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        _xw,_yw = self.xmw,self.ymw
        _tc = int(self.dpi /25.4*tick)  # mm -> pix
        for i in range(self.nim):
            self.imCur[i] = self.imOvl_Cursor(self.im[i],_xw,_yw,_tc,_tc)

    def imRGB_all(self, mode:int=0):
        """   im[] -> imR,G,B[] RGB-band separated images 
        mode = 0 : imR,G,B[].mode == "L" 8-bit B/W
        mode = 1 :   == "RGB" 3x8-bit true color (R,0,0) (0,G,0) (0,0,B)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        for i in range(self.nim):
            self.imR[i],self.imG[i],self.imB[i] = \
                self.imRGB(self.im[i],mode,fill=64)  # fill=0だと暗いので
    
    def imMask_all(self):
        """  imR,G,B[] -> imRm,Gm,Bm[] Circle-Masked images
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        for i in range(self.nim):
            self.imRm[i] = self.imMask_Ellipse(\
                    self.imR[i], self.xmw,self.ymw)        
            self.imGm[i] = self.imMask_Ellipse(\
                    self.imG[i], self.xmw,self.ymw)        
            self.imBm[i] = self.imMask_Ellipse(\
                    self.imB[i], self.xmw,self.ymw)        

    def imOD_all(self):
        """  imRm,Gm,Bm[] -> odR,G,B[]: ndarray
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        for i in range(self.nim):
            self.odR[i], self.imRod[i] = self.imOD(self.imRm[i])
            self.odG[i], self.imGod[i] = self.imOD(self.imGm[i])
            self.odB[i], self.imBod[i] = self.imOD(self.imBm[i])
 
    def show_all(self,mode:int=4, plt=False) -> Image :
        """  Show images
        mode = 0 : show im0
             = 1 :  im[]
             = 2 :  imCur[]
             = 3 :  imR,G,B[]
             = 4 :  imRm,Gm,Bm[] Masked image
             = 5 :  imRod,God,Bod[] Optical Density image
        plt: bool   SW for plot image
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        im = self.im[0]
        if(im is None): return
        imw,imh = im.size
        n = self.nim   # ima size は、6枚用と同じにしておく
        
        def _paste(imR,imG,imB,color=(0,0,0)) -> Image:
            ima = Image.new('RGB',((imw+10)*n,(imh+10)*3),color)
            for i in range(self.nim):          # ->r#01)
                ima.paste(imR[i],((imw+10)*i+5,5))
                ima.paste(imG[i],((imw+10)*i+5,(imh+10)*1+5))
                ima.paste(imB[i],((imw+10)*i+5,(imh+10)*2+5))
            return(ima)

        if(mode == 0):
            ima = self.im0
        if(mode == 1):
            ima = Image.new('RGB',((imw+10)*n,imh+10))
            for i in range(self.nim):
                ima.paste(self.im[i],((imw+10)*i+5,5))
        if(mode == 2):
            ima = Image.new('RGB',((imw+10)*n,imh+10))
            for i in range(self.nim):
                ima.paste(self.imCur[i],((imw+10)*i+5,5))
        if(mode == 3):
            ima = _paste( self.imR, self.imG, self.imB )
        if(mode == 4):
            ima = _paste( self.imRm, self.imGm, self.imBm )
        if(mode == 5):
            ima = _paste( self.imRod, self.imGod, self.imBod,\
                         color=(255,255,255) )
        if(plt): ima.show()
        return( ima )

    def save_all(self,fnH:str='_', modes:list=[2,12,42]):
        """  Save images & O.D. array
        fnH:    str   file name header
        modes:  [int,..]    Job modes -> see below
        mode 
        = 1 :         im[]                  save as '.png'
        = 2 :         imCur[]                       '.png'
        = 11,12,13 :  imR,G,B[]                     '.png'
        = 14,15,16 :  imR,G,B[]                     '.csv'
        = 21,22,23 :  imRm,Gm,Bm[] Masked image     '.png'
        = 31,32,33 :  imRod,God,Bod[] O.D.image     '.csv'
        = 41,42,43 :  odR,G,B[] O.D. 2Darray        '.csv'  
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        from string import Template   # ->r#06)

        job = []
        #  mode                        typ (1:im->png, 2:np->csv, 3:im->csv)
        if(  1 in modes):   job.append([1,'im[$no]',self.im])
        if(  2 in modes):   job.append([1,'imCur[$no]',self.imCur])
        if( 11 in modes):   job.append([1,'imR[$no]',self.imR])
        if( 12 in modes):   job.append([1,'imG[$no]',self.imG])
        if( 13 in modes):   job.append([1,'imB[$no]',self.imB])
        if( 14 in modes):   job.append([3,'imR[$no]',self.imR])
        if( 15 in modes):   job.append([3,'imG[$no]',self.imG])
        if( 16 in modes):   job.append([3,'imB[$no]',self.imB])
        if( 21 in modes):   job.append([1,'imRm[$no]',self.imRm])
        if( 22 in modes):   job.append([1,'imGm[$no]',self.imGm])
        if( 23 in modes):   job.append([1,'imBm[$no]',self.imBm])
        if( 24 in modes):   job.append([3,'imRm[$no]',self.imRm])
        if( 25 in modes):   job.append([3,'imGm[$no]',self.imGm])
        if( 26 in modes):   job.append([3,'imBm[$no]',self.imBm])
        if( 31 in modes):   job.append([1,'imRod[$no]',self.imRod])
        if( 32 in modes):   job.append([1,'imGod[$no]',self.imGod])
        if( 33 in modes):   job.append([1,'imBod[$no]',self.imBod])
        if( 41 in modes):   job.append([2,'odR[$no]',self.odR])
        if( 42 in modes):   job.append([2,'odG[$no]',self.odG])
        if( 43 in modes):   job.append([2,'odB[$no]',self.odB])
        for i,_job in enumerate(job):
            typ,fmt,im = _job
            s = Template(fmt)
            for j in range(self.nim): 
                fn = s.substitute(no=j)
                if(typ == 1):   
                    im[j].save(fnH+fn+'.png')
                if(typ == 2):   
                    np.savetxt(fnH+fn+'.csv',im[j],\
                      fmt='%.3f',delimiter=',') 
                if(typ == 3):   
                    nda = np.array( im[j], dtype=np.uint8 )
                    np.savetxt(fnH+fn+'.csv',nda,\
                      fmt='%d',delimiter=',') 

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class rdGafScan6_Kam
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class rdGafScan6_Kam( rdGafScan6 ):
    """ Read Gaf Scan Image (E5A用 Gaf□85x85 6枚貼 神原Scan用)
Params. for constructor:
    fnScn:  str         Gaf scan image file name
    dpi:    int         [dpi] defa=75 
    ana:    bool        SW for scnAna_all() 
Attributes:
    see super()._myclear()
HowTo:
    see class:rdGafScan6()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
                      
    def __init__(self, fnScn:str, dpi=75, ana=True ):
        """  Constructor
        fnScn:  str         Gaf scan image file name
        dpi:    int         [dpi] defa=75 
            cf.) now support dpi=75,300 only
        ana:    bool        SW for scnAna_all() 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        super().__init__(fnScn,dpi,ana=False)
        """ w10:ペイントで表示してカーソル位置で確認する！ """
        if(dpi == 75):
            self.x0     = [40,320]
            self.y0     = [20,300,585]
            self.xw,self.yw =   250,250 
            self.xmw,self.ymw = 210,210
        if(dpi == 300):     #""" not measured yet """
            mag= 300./75.
            self.x0     = [150,1325]
            self.y0     = [200,1300,2400]
            self.xw,self.yw =   250*mag,250*mag
            self.xmw,self.ymw = 210*mag,210*mag
        self._imCropPrep()
        if(ana):
            self.scnAna_all()


""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class rdGafScan1
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class rdGafScan1( rdGafScan6 ):
    """ Read Gaf Scan Image (E5A用 Gaf□85x85 1枚の読取り)
Params. for constructor:
    fnScn:  str         Gaf scan image file name
    dpi:    int         [dpi] defa=75 
Attributes:
    see ._myclear()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
                      
    def __init__(self, fnScn:str, dpi=75, ana=True):
        """  Constructor
        fnScn:  str         Gaf scan image file name
        dpi:    int         [dpi] defa=75 
        ana:    bool        SW for scnAna_all() 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        super().__init__(fnScn,dpi,ana=False)
        """ w10:ペイントで表示してカーソル位置で確認する！ """
        self.nim    = 1
        self.x0     = [0,0]
        self.y0     = [0,0,0]
        self.xw,self.yw =   250,250 
        self.xmw,self.ymw = 210,210
        if(dpi == 300):     #""" not measured yet """
            mag= 300./75.
            self.xw,self.yw =   250*mag,250*mag
            self.xmw,self.ymw = 210*mag,210*mag
        self._imCropPrep()
        if(ana):
            self.scnAna_all()

    def imSep_all(self):        # override 
        """  im[0] = im0.crop : no separation
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def _imSep1(ix, iy) -> Image:
            xp,yp = self.x0[ix], self.y0[iy]
            im1 = self.imCrop(self.im0,xp,yp,self.xw,self.yw)
            return( im1 )

        self.im[0] = _imSep1(0,0)


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest01 =  True    # test rdGafScn6()
    SwTest02 =  False   # test rdGafScn6_Kam()
    SwTest03 =  False   # test rdGafScn1()

    SwTest11 =  False   # test scnShift()
    SwTest12 =  False   # test scnShift_all()

    SwTest91 =  False   # test scnSave()
    """-.---.---.---.---.---.---.---.---.---.---.---.---"""
    fndir = '../0000Test/_GafData/'   # ExpData folder path

    if(SwTest01 or SwTest02):
        fn_scn,dpi = fndir+'testGaf6ay_75dpi.bmp', 75     # for test
        #sft = [None]*6
        #fn_scn,dpi = fndir+'testGaf6ay_300dpi.bmp', 300   # for test
        #sft = [None]*6
        #fn_scn,dpi = fndir+'2105Kr,Ar_75dpi.bmp', 75
        #sft = [(0,-3)]*6
        sft = [None]*6
        scn = rdGafScan6(fn_scn,dpi=dpi)
    if(SwTest02):
        fn_scn,dpi = fndir+'testGaf6km_75dpi.bmp', 75     # for test
        sft = [(-1,-2),(-2,-2),(-2,-2),(-2,-2),(-2,-2),(-2,-2)]
        scn = rdGafScan6_Kam(fn_scn,dpi=dpi)
    if(SwTest03):
        fn_scn,dpi = fndir+'testGaf1_75dpi.bmp', 75
        sft = [(-2,-2)]
        scn = rdGafScan1(fn_scn,dpi=dpi)

    """ test01～test12 : test scn.show_all()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(True):
        scn.show_all(mode=0,plt=True)   # show im0           
        scn.show_all(mode=1,plt=True)   # show im[]
        scn.show_all(mode=2,plt=True)   # show imCur[]
        scn.show_all(mode=3,plt=True)   # show imR,G,B[] RGB bands
        scn.imRGB_all(mode=1)           # RGB colored image
        scn.show_all(mode=3,plt=True) 
        scn.show_all(mode=4,plt=True)   # show imRm,Gm,Bm[] Masked image
        scn.show_all(mode=5,plt=True)   # show imRod,God,Bod[] Optical Density

    """ test01～test12 : test scn.save_all()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(False):
        scn.save_all(modes=[31,41])        
        
    """ test11 : test scn.scnShift()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest11):
        for i,_sft in enumerate(sft):
            if(_sft is None): continue
            dx,dy=_sft
            scn.scnShift(i,dx,dy)
        scn.scnAna_all()
        scn.show_all(mode=2,plt=True)

    """ test12 : test scn.scnShift_all()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest12):
        if(sft[0] is not None):
            dx,dy=sft[0]
            scn.scnShift_all(dx,dy)
        scn.show_all(mode=2,plt=True)

    """ test91 : test scn.scnSave_all()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest91):
        scn.save_all()
        #scn.save_all(modes=[1,2,11,21,31,41])


""" ========================================================
        ref) Numbered ref
    ========================================================
r#01) note.nkmk.me　Python, Pillowで画像を縦・横に連結（結合）
    https://note.nkmk.me/python-pillow-concat-images/

r#02) note.nkmk.me　Python, Pillowで二枚の画像をマスク画像に従って合成
    https://note.nkmk.me/python-pillow-composite/ 

      PIL doc. Tutrial: Processing individual bands
r#03) Image class
    https://pillow.readthedocs.io/en/stable/index.html

r#04) Tutrial: Processing individual bands 
    https://pillow.readthedocs.io/en/stable/handbook/tutorial.html

r#05) よちよちpython RGBをRGBAに変換する方法と画像合成
    https://chayarokurokuro.hatenablog.com/entry/2020/08/14/113624

r#06) 6.1.4. テンプレート文字列
    https://docs.python.org/ja/3.5/library/string.html

        ref) Others
    ======================================================== 
●note.nkmk.me 画像処理
　https://note.nkmk.me/image-processing/
    
A4 = X:210 x Y:297 mm 
mm / 25.4 x dpi = px

------------------------------------------------------------------
"""